unit ckemu;

interface

uses
  Windows,
  SysUtils,
  Classes,
  ckfixelf;


function compilewithdcc(const ckdir:string; const parameters:string; const encoding:string; write:TCompilerWriteCallback; outputstream: TStream; debug:boolean):string;

implementation


type
  TProcessInfo = record
    syscallSP: Cardinal;
    userSP: Cardinal;
    root_pid: integer;
    start_brk, brk, start_code, end_code, end_data, start_stack: Cardinal;
    arg_start, arg_end, env_start, env_end: Cardinal;
    Debugger: Integer;
    FixedMem: array[0..16383] of byte;
  end;
  PProcessInfo = ^TProcessInfo;

function compilewithdcc(const ckdir:string; const parameters:string; const encoding:string; write:TCompilerWriteCallback; outputstream: TStream; debug:boolean):string;
var
  SI: TStartupInfo;
  PI: TProcessInformation;
  createflags: Cardinal;
  DebugEvent: TDebugEvent;
  env: string;
  WasOk: boolean;
  ProcessHandle: THandle;
  ThreadHandle: THandle;
  ThreadContext: TContext;
  Instruction: word;
  BytesRead: Cardinal;
  SyscallHandler: Cardinal;
  ProcessInfo: PProcessInfo;
  AllocMem: Cardinal;
  AllocRet: Cardinal;
  AllocInfo: TMemoryBasicInformation;
  EmptyMemory: array[0..65535] of byte;
  c: byte;
  OutputString: string;
  dcc: string;
  ckdir_forward:string;
  ckdir_faked:string;
  cdir:string;
  seek_res:int64;

  procedure writeln(const s:string);
  begin
    write(s+#13#10);
  end;

begin
  result:='';
  fillchar(EmptyMemory, 65536, 0);
  processinfo:=nil;
  ThreadHandle:=0;
  ProcessHandle:=0;
  SyscallHandler:=0;

  //ckdir:=includetrailingpathdelimiter(extractfilepath(paramstr(0)));
  ckdir_forward:=StringReplace(ckdir,'\','/',[rfReplaceAll]);

  if pos(':\',ckdir)<>2 then
  begin
    writeln('You seem to be running this from some strange drive mapping.');
    writeln('CrossKylix needs a real drive (e.g. c:\) to work.');
    exit;
  end;

  ckdir_faked:=ckdir_forward;
  ckdir_faked[2]:=ckdir_faked[1];
  ckdir_faked[1]:='/';

  cdir:=ExtractShortPathName(includetrailingpathdelimiter(getcurrentdir));
  dcc:=ckdir+'kylix\bin\dcc';

  if not fileexists(dcc) then
  begin
    writeln('Unable to find dcc at '+dcc);
    exit;
  end;

  dcc:=StringReplace(dcc,'\','/',[rfReplaceAll]);

  with SI do
  begin
    FillChar(SI, SizeOf(SI), 0);
    cb := SizeOf(SI);
    dwFlags := STARTF_USESHOWWINDOW or STARTF_USESTDHANDLES;
    wShowWindow := SW_HIDE;
    hStdInput := GetStdHandle(STD_INPUT_HANDLE);
    hStdOutput := GetStdHandle(STD_OUTPUT_HANDLE);
    hStdError := GetStdHandle(STD_ERROR_HANDLE);
  end;

  createflags := 0;
  createflags := createflags or DEBUG_PROCESS or PROCESS_VM_READ or PROCESS_VM_WRITE;
  env := 'HOME='+ckdir_faked + #0 + 'LD_LIBRARY_PATH='+ckdir_faked+'lib:'+ckdir_faked+'kylix/bin' + #0 + 'LC_CTYPE='+encoding + #0 + 'LANG=POSIX' + #0#0;

  if debug then
  begin
    writeln('Env: '+env);
    writeln('Running: '+ckdir+'ck.dat -f -w'+getcurrentdir+' -c'+ckdir_forward+' "'+dcc+'" "'+dcc+' '+parameters+'"');
  end;

  WasOK := CreateProcess(nil, PChar(ckdir+'ck.dat -f -w'+cdir+' -c'+ckdir_forward+' "'+dcc+'" "'+dcc+' '+parameters+'"'), nil, nil, True,
    createflags, @env[1], pchar(ckdir), SI, PI);


  if not WasOK then
  begin
    writeln('Error running DCC');
    exit;
  end;

  ThreadContext.ContextFlags := CONTEXT_CONTROL or CONTEXT_INTEGER;

  while WaitForDebugEvent(DebugEvent, INFINITE) do
  begin
  //  if debug then writeln('DEBUG EVENT:'+inttostr(DebugEvent.dwDebugEventCode));
    case DebugEvent.dwDebugEventCode of
      CREATE_PROCESS_DEBUG_EVENT:
        begin
          ProcessHandle := DebugEvent.CreateProcessInfo.hProcess;
          ThreadHandle := DebugEvent.CreateProcessInfo.hThread;
          if debug then
          begin
            writeln('Got process handle ' + inttostr(ProcessHandle));
            writeln('Got thread handle ' + inttostr(ThreadHandle));
          end;
        end;
      LOAD_DLL_DEBUG_EVENT:
        begin
          // we don't really care about loaded DLLs.
        end;
      EXCEPTION_DEBUG_EVENT:
        begin
          // ignore Breakpoints
          // if debug then writeln('Exception: '+inttostr(DebugEvent.Exception.ExceptionRecord.ExceptionCode));
          if DebugEvent.Exception.ExceptionRecord.ExceptionCode <> EXCEPTION_BREAKPOINT then
          begin
            if GetThreadContext(ThreadHandle, ThreadContext) then
            begin
              // if debug then writeln('ThreadContext.EIP: '+inttostr(ThreadContext.EIP));
              if ReadProcessMemory(ProcessHandle, pointer(ThreadContext.EIP), @Instruction, 2, BytesRead) then
              begin
                if Instruction = $80CD then // syscall
                begin
                  ThreadContext.Eip := ThreadContext.Eip + 2;
                  if ThreadContext.EAX = $DEADBEEF then
                  begin
                    if debug then writeln('Get syscall handler');
                    SyscallHandler := ThreadContext.EBX;
                    ProcessInfo := PProcessInfo(pointer(ThreadContext.ECX));
                    if SetThreadContext(ThreadHandle, ThreadContext) then
                    begin
                    end
                    else
                    begin
                      writeln('Unable to write thread context');
                    end;
                  end
                  else
                  begin
                    if debug then writeln('Syscall '+inttostr(ThreadContext.EAX));

                    if (ThreadContext.EAX = 4) and (ThreadContext.EBX < 3) then
                    begin // this is console output.
                      setlength(OutputString, ThreadContext.EDX);
                      if ReadProcessMemory(ProcessHandle, pointer(ThreadContext.ECX), @OutputString[1], ThreadContext.EDX, BytesRead) then
                      begin
                        write(OutputString);
                        ThreadContext.EAX:=BytesRead;
                      end
                      else
                      begin
                        writeln('Unable to get dcc output');
                      end;
                    end
                    else
                    if (ThreadContext.EAX = 5) and (ThreadContext.ECX = 577) then
                    begin // this is our created binary file, give it a fake fd
                      setlength(OutputString, 255);
                      if ReadProcessMemory(ProcessHandle, pointer(ThreadContext.EBX), @OutputString[1], 255, BytesRead) then
                      begin
                        if debug then writeln('Output: '+pchar(OutputString)+' Flags: '+inttostr(ThreadContext.ECX));
                        result:=pchar(OutputString);
                        ThreadContext.EAX:=4333;
                      end
                      else
                      begin
                        writeln('Unable to get dcc output');
                      end;
                    end
                    else
                    if (ThreadContext.EAX = 4) and (ThreadContext.EBX = 4333) then
                    begin // write on our fake fd
                      setlength(OutputString, ThreadContext.EDX);
                      if ReadProcessMemory(ProcessHandle, pointer(ThreadContext.ECX), @OutputString[1], ThreadContext.EDX, BytesRead) then
                      begin
                        outputstream.Write(outputstring[1],BytesRead);
                        ThreadContext.EAX:=BytesRead;
                      end
                      else
                      begin
                        writeln('Unable to get dcc binary output');
                      end;
                    end
                    else
                    if (ThreadContext.EAX = 108) and (ThreadContext.EBX = 4333) then
                    begin // stat on our fake fd. Ignore it, dcc seems to survive this.
                      ThreadContext.EAX:=cardinal(-1);
                    end
                    else
                    if (ThreadContext.EAX = 6) and (ThreadContext.EBX = 4333) then
                    begin // close on our fd.
                      ThreadContext.EAX:=0;
                    end
                    else
                    if (ThreadContext.EAX = 140) and (ThreadContext.EBX = 4333) then
                    begin // llseek on our fd. assume we'll never have files > 4gb. and throw away the high part :)
                      if ThreadContext.EDI=0 then
                      begin // SEEK_SET
                        outputstream.position:=ThreadContext.EDX;
                      end
                      else
                      if ThreadContext.EDI=1 then
                      begin // SEEK_CUR
                        outputstream.position:=outputstream.position+ThreadContext.EDX;
                      end
                      else
                      if ThreadContext.EDI=1 then
                      begin // SEEK_END
                        outputstream.position:=outputstream.Size-ThreadContext.EDX;
                      end;
                      seek_res:=outputstream.position;

                      if not WriteProcessMemory(ProcessHandle, pointer(ThreadContext.ESI), @seek_res, 8, BytesRead) then
                      begin
                        writeln('Error while seeking in binary file.');
                      end;

                      ThreadContext.EAX:=0;
                    end
                    else
                    begin
                      if (ThreadContext.EBX = 4333) then
                      begin // noone else should use this FD.
                        writeln('Warning: Probably unhandled syscall for binary output. '+inttostr(ThreadContext.EAX));
                      end;

                      if debug and (ThreadContext.EAX = 5) then
                      begin
                        setlength(OutputString, 255);
                        if ReadProcessMemory(ProcessHandle, pointer(ThreadContext.EBX), @OutputString[1], 255, BytesRead) then
                        begin
                          writeln('Open: '+pchar(OutputString)+' Flags: '+inttostr(ThreadContext.ECX));
                        end;
                      end;

                      // push EIP on the stack for RET
                      ThreadContext.Esp := ThreadContext.Esp - 4;
                      if WriteProcessMemory(ProcessHandle, pointer(ThreadContext.Esp), @ThreadContext.Eip, 4, BytesRead) then
                      begin
                        ThreadContext.Eip := SyscallHandler;
                      end
                      else
                      begin
                        writeln('Unable to push EIP onto stack');
                      end;
                    end;

                    if not SetThreadContext(ThreadHandle, ThreadContext) then
                    begin
                      writeln('Unable to write thread context');
                    end;
                  end;
                end
                else
                begin
                  if DebugEvent.Exception.ExceptionRecord.ExceptionCode = EXCEPTION_ACCESS_VIOLATION then
                  begin
                    AllocMem := DebugEvent.Exception.ExceptionRecord.ExceptionInformation[1];
                    if debug then writeln('Allocmem: '+inttostr(allocmem));
                    AllocMem := AllocMem and $FFFF0000; // 64k align
                    AllocRet := VirtualQueryEx(ProcessHandle, pointer(AllocMem), AllocInfo, sizeof(AllocInfo));
                    if AllocRet > 0 then
                    begin
                      if (allocInfo.AllocationProtect = PAGE_EXECUTE_READWRITE) and (allocInfo.State = MEM_RESERVE) then
                      begin
                        allocRet := Cardinal(VirtualAllocEx(ProcessHandle, pointer(AllocMem), $FFFF, MEM_COMMIT, PAGE_EXECUTE_READWRITE));
                        if allocmem <> allocret then
                        begin
                          writeln('VirtualAllocEx failed');
                        end
                        else
                        begin
                          if WriteProcessMemory(ProcessHandle, pointer(allocmem), @EmptyMemory, $FFFF, AllocRet) and (AllocRet = $FFFF) then
                          begin
                            if ReadProcessMemory(ProcessHandle, @ProcessInfo.FixedMem[AllocMem shr 16], @c, 1, BytesRead) then
                            begin
                              if c = 2 then
                              begin
                                c := 1;
                                if WriteProcessMemory(ProcessHandle, @ProcessInfo.FixedMem[AllocMem shr 16], @c, 1, BytesRead) then
                                begin
                                  // done
                                end
                                else
                                begin
                                  writeln('Error writing process info');
                                end;
                              end
                              else
                              begin
                                writeln('Process info is wrong');
                              end;
                            end
                            else
                            begin
                              writeln('Error reading process info');
                            end;
                          end
                          else
                          begin
                            writeln('Error initalizing memory');
                          end;
                        end;
                      end;
                    end
                    else
                    begin
                      writeln('VirtualQueryEx failed');
                    end;
                  end;
                end;
              end
              else
              begin
                writeln('Unable to read process memory');
              end;
            end
            else
            begin
              writeln('Unable to get thread context');
            end;
          end;
        end;
      EXIT_PROCESS_DEBUG_EVENT:
        begin
          ContinueDebugEvent(DebugEvent.dwProcessId, DebugEvent.dwThreadId, DBG_CONTINUE);
          break;
        end;
    end;
    ContinueDebugEvent(DebugEvent.dwProcessId, DebugEvent.dwThreadId, DBG_CONTINUE);
  end;
  CloseHandle(PI.hThread);
  CloseHandle(PI.hProcess);
end;

end.


