unit ckfixelf;

{$WARN SYMBOL_PLATFORM OFF}
interface

type
  TCompilerWriteCallback = procedure(const s:string) of object;

procedure fixsegments(location: Pointer; write:TCompilerWriteCallback; debug: boolean);

implementation

uses
  SysUtils;

type
  Elf32_Addr = Cardinal;
  Elf32_Half = Word;
  Elf32_Off = Cardinal;
  Elf32_Sword = Integer;
  Elf32_Word = Cardinal;

  TElfHeader = packed record
    e_ident: packed array[0..15] of char;
    e_type: Elf32_Half;
    e_machine: Elf32_Half;
    e_version: Elf32_Word;
    e_entry: Elf32_Addr;
    e_phoff: Elf32_Off;
    e_shoff: Elf32_Off;
    e_flags: Elf32_Word;
    e_ehsize: Elf32_Half;
    e_phentsize: Elf32_Half;
    e_phnum: Elf32_Half;
    e_shentsize: Elf32_Half;
    e_shnum: Elf32_Half;
    e_shstrndx: Elf32_Half;
  end;

  PElfHeader = ^TElfHeader;

  TProgramHeader = packed record
    p_type: Elf32_Word;
    p_offset: Elf32_Off;
    p_vaddr: Elf32_Addr;
    p_paddr: Elf32_Addr;
    p_filesz: Elf32_Word;
    p_memsz: Elf32_Word;
    p_flags: Elf32_Word;
    p_align: Elf32_Word;
  end;

  PProgramHeader = ^TProgramHeader;

procedure fixsegments(location: Pointer; write:TCompilerWriteCallback; debug: boolean);
var
  ElfHeader: PElfHeader;
  i: integer;
  ProgramHeaders: array of PProgramHeader;
  ProgramHeaderOffset: integer;

  procedure writeln(const s:string);
  begin
    write(s+#13#10);
  end;

begin
  ElfHeader := location;

  if debug then
  begin
    if (ElfHeader.e_ident[1] <> 'E') or (ElfHeader.e_ident[2] <> 'L') or (ElfHeader.e_ident[3] <> 'F') then
    begin
      writeln('Error: This is not a valid linux ELF binary.');
      halt(3);
    end;

    write('Checking ELF Header... ');

    if ElfHeader.e_phentsize = sizeof(TProgramHeader) then
    begin
      writeln('OK');
    end
    else
    begin
      writeln('');
      writeln('Error: Unsupported Program Header size.');
      writeln('If this file really is a Kylix compiled ELF executable, please report');
      writeln('back to the author.');
      halt(4);
    end;

    writeln('Reading Program Headers...');
  end;

  setlength(ProgramHeaders, ElfHeader.e_phnum);
  ProgramHeaderOffset := ElfHeader.e_phoff;

  for i := 0 to ElfHeader.e_phnum - 1 do
  begin
    ProgramHeaders[i] := pointer(ProgramHeaderOffset + integer(location) + i * sizeof(TProgramHeader));
  end;

  if debug then
  begin
    writeln('Checking Program Header table...');
  end;

  for i := 0 to ElfHeader.e_phnum - 1 do
  begin
    if ProgramHeaders[i].p_type = 1 then
    begin
      if debug then
      begin
        write('  Loadable segment at ' + inttohex(ProgramHeaders[i].p_vaddr, 8) + '/' + inttohex(ProgramHeaders[i].p_paddr, 8) + ', Size ' + inttohex(ProgramHeaders[i].p_memsz, 8) + ', Flags ');
        if (ProgramHeaders[i].p_flags and 4) = 4 then write('R');
        if (ProgramHeaders[i].p_flags and 2) = 2 then write('W');
        if (ProgramHeaders[i].p_flags and 1) = 1 then write('E');
        writeln('');
      end;
      if (ProgramHeaders[i].p_flags and 1) <> 1 then
      begin
        if debug then
        begin
          writeln('  Missing executable flag, fixing... ');
        end;
        ProgramHeaders[i].p_flags:=ProgramHeaders[i].p_flags or 1;
      end;
    end;
  end;
end;

end.

