// CrossKylix IDE Plugin, (c) Simon "scamp" Kissel 2004. http://crosskylix.untergrund.net
// This code is licensed under the BSD license. For more information see the CrossKylix website
// If you improve this code, PLEASE send back your changes to me (scamp@untergrund.net)

{

0.9.1   10/08/2004

- A bug in ekemu.pas made CrossKylix fail when installed in directory with a space in it.
  Thanks to Alan for finding this bug.
- When compiling from inside the Delphi IDE, the ck.dat process wasn't freed. Thanks to
  Ghislain for the notification.

0.9.0   10/07/2004

- Finished a complete rewrite of the emulation engine. Everything now is native Win32 code,
  LINE is gone, and that crappy cygwin1.dll isn't used anymore \o/
- The outer part of the emulation engine has been moved to Delphi code, and now resides
  in the ckdcc.pas unit. The only stuff still written in C is the Linux ELF loader code and
  some syscalls implementations.
- Changed form positions to poScreenCenter for better multi-monitor support (thanks Craig)
- Added the option to output a full trace of all linux syscalls done by DCC. This is for
  debugging purposes, and hell slow.
- A command-line compiler, ckdcc, is now part of the project
- Added all the missing locale-stuff. CK is now able to compile source files that use characters
  outside the ANSI set, e.g. german umlauts. In the CK Options dialog it's now possible to
  select the source encoding. The default, ISO-8859-15, should be fine for western europe
  and the USA.
- As reported by Kristijan (Thanks!), my Libc dlopen workaround did not work on Debian distros.
  This is now fixed.
- As recently reported, Kylix compiled applications crash on Linux kernels with exec-shield
  enabled. I've provided a fix for this at http://crosskylix.untergrund.net/execshield/.
  This fix now automaticly is applied in-memory on CrossKylix-compiled executables, so you
  are no longer affected by this Kylix compiler issue.

  Version history:

0.8.2   07/30/2004

- Fixed a bug that was causing CrossKylix to hang if dcc did output a lot of error messages
  while compiling a project
- The installkylix.exe installer did not install optional packages like dbexpress
- Now uses the GetShortPathName API call for getting short path names for paths that
  contain spaces. CrossKylix should now work for projects with spaces in path names.
  Also installing CrossKylix to a directory with spaces in it should be fine now.
  Thanks to Jim Mckeeth for the idea.
- The strange cygwin1.dll bug that requires a lot of CK users to do a minimal Cygwin
  installation still isn't found or fixed. Sigh.

0.8.1   07/06/2004

- Finally found out what was causing the Access Violations when closing Delphi
  and also caused the missing Tools menu item after restart - this is caused by
  adding the CrossKylix menu item at the end of the tools menu. Sigh.
  Now the item is added in the upper part of the tools menu, and it works.
  Open Tools Api sometimes can be a real mystery...
- Added $WARN UNSAFE_TYPE OFF/$WARN UNSAFE_CODE OFF for D7
- Project files with "unit1 in 'unit1.pas'" had problems finding unit1.pas,
  as Delphi/Kylix does not use the unit search path in this case, but uses
  the current directory. WHY? Added work-around to line/linexec
- If dcc inside CrossKylix crashed for some reason, it kept stuck in an endless-
  loop causing also the Delphi IDE to freeze. It should now abort correctly in this
  case

0.8     07/03/2004

- Now uses GetCurrentProject to find out active project.
- Sets LANG=en_US
- Added Andreas Hausladen's ProcessAffinity stuff to make LINE SMP-compatible
- Now adds Kylix bin path to LD_LIBRARY_PATH so that libtextform.so can be used (so that
  text xfm/dfm files work with CrossKylix)
- Tons of fixes to the line/linexec stuff to workaround a cygwin bug preventing the usage of libtextform.so
- Now uses better buffering for stderr and stdout, resulting in no more messed up error/warning messages
- Fixed a strange bug that was causing the stack to overrun, resulting in CrossKylix silently stopping to compile
  for a lot of projects (including all CLX stuff)
- If exe and unit output directorys are not configured in project options, add project dir as default
  (no more "where is my exe?" bug reports ;)
- Added the option to use a patched Libc.pas, which loads .so's from the application directory = much less
  hazzle deploying kylix applications
- Fixed lots of bugs. Surprise, surprise.
- Fixed a bug that sometimes gave an Access Violation when removing/recompiling the IDE plugins

0.7     06/08/2004

- Removed int80.sys
- Removed patchdcc.exe
- Added path translations to linexec
- Fatal errors are now displayed as errors in compile dialog
- Compile dialog is now StayOnTop
- Error box if unable to create configuration
- Forced CrossKylix path to include a trailing backslash
- Fixed problem that could result in an AV instead of displaying the
CrossKylix options menu item under Tools
- Added cfg filters for new Delphi 7 compiler directives not supported
by Kylix (please report if there are more)

}

unit ckmain;

interface
procedure Register;
procedure dolog(const msg: string);

implementation
uses
  ToolsAPI, Dialogs, Windows, menus, SysUtils, winsock, classes, controls, forms, ComCtrls, Graphics, inifiles, ckcompiledialog, ckoptionsdialog, ckemu, ckfixelf;

type

  TCrossKylixAddIn = class(TNotifierObject, IOTAWIzard)
  private
    ProjectMenu: TMenuItem;
    ToolsMenu: TMenuItem;
    CompileMenuItem: TMenuItem;
    BuildMenuItem: TMenuItem;
    SeparatorMenuItem: TMenuItem;
    ToolsMenuItem: TMenuItem;
    cd: TCompileDialog;
    od: TOptionsDialog;
    ini: TMemInifile;
    LinuxBaseDir, BaseDir, UnitDirectories, ObjectDirectories, IncludeDirectories: string;
    Encoding: string;
    DebugOutput: boolean;
    DebugTrace: boolean;
    UsePatchedLibc: boolean;
    Line: string;
    projectpath: string;
    currentdir: string;
    w_count, e_count, h_count, l_count: integer;
    currentfile: string;
    cc: integer;
    function GetIDString: string;
    function GetName: string;
    function GetState: TWizardState;
    procedure execute;
    procedure docompile(build: boolean);
    procedure msgoutput(const s: string);
  public
    procedure buildkylix(Sender: TObject);
    procedure compilekylix(Sender: TObject);
    procedure showoptions(Sender: TObject);
    constructor Create;
    destructor Destroy; override;
  end;

var
  GetProcessAffinityMask: function(hProcess: THandle;
    var lpProcessAffinityMask, lpSystemAffinityMask: DWORD): BOOL; stdcall;
  SetProcessAffinityMask: function(hProcess: THandle;
    dwProcessAffinityMask: DWORD): BOOL; stdcall;
  DefaultProcessAffinity: Cardinal;

  // This code is provided by Andreas Hausladen

procedure SetAffinity;
var
  Lib: HMODULE;
  SysAffinity: Cardinal;
begin
  Lib := GetModuleHandle(kernel32);
  @GetProcessAffinityMask := GetProcAddress(Lib, 'GetProcessAffinityMask');
  @SetProcessAffinityMask := GetProcAddress(Lib, 'SetProcessAffinityMask');
  if Assigned(SetProcessAffinityMask) then
  begin
    GetProcessAffinityMask(GetCurrentProcess, DefaultProcessAffinity, SysAffinity);
    SetProcessAffinityMask(GetCurrentProcess, 1);
    Sleep(100);
  end;
end;

procedure ResetAffinity;
begin
  if Assigned(SetProcessAffinityMask) then
    SetProcessAffinityMask(GetCurrentProcess, DefaultProcessAffinity);
end;

function GetMainFilename(module: IOTAModule):string;
var i: integer;
    editor: IOTAEditor;
    fname: string;
    ext: string;
begin
  result:='';
  for i:=0 to module.ModuleFileCount-1 do
  begin
    editor:=module.ModuleFileEditors[i];
    if assigned(editor) then
    begin
      fname:=editor.FileName;
      ext:=lowercase(ExtractFileExt(fname));
      if (ext='.dpr') or (ext='.dpk') then
      begin
        result:=fname;
      end;
    end;
  end;
end;

function GetCurrentProject: IOTAProject;
var
  Services: IOTAModuleServices;
  Module: IOTAModule;
  Project: IOTAProject;
  ProjectGroup: IOTAProjectGroup;
  MultipleProjects: Boolean;
  I: Integer;
begin
  Result := nil;
  MultipleProjects := False;
  Services := BorlandIDEServices as IOTAModuleServices;
  for I := 0 to Services.ModuleCount - 1 do
  begin
    Module := Services.Modules[I];
    if Module.QueryInterface(IOTAProjectGroup, ProjectGroup) = S_OK then
    begin
      Result := ProjectGroup.ActiveProject;
      Exit;
    end
    else if Module.QueryInterface(IOTAProject, Project) = S_OK then
    begin
      if Result = nil then
        // Found the first project, so save it
        Result := Project
      else
        MultipleProjects := True;
      // It doesn't look good, but keep searching for a project group
    end;
  end;
  if MultipleProjects then
    Result := nil;
end;

function FindClassForm(const AClassName: string): TForm;
var
  I: Integer;
begin
  Result := nil;
  with Screen do
    for I := 0 to FormCount - 1 do
      if Forms[I].ClassNameIs(AClassName) then
      begin
        Result := Forms[I];
        Break;
      end;
end;

function FindMessageViewForm: TForm;
begin
  // if TMsgWindow exists all messages are sent to it
  Result := FindClassForm('TMsgWindow');
  if not Assigned(Result) then // otherwise TMessageViewForm is used
    Result := FindClassForm('TMessageViewForm');
end;

procedure ShowMessageView(ScrollToLast: Boolean = False);
var
  MessageViewForm: TForm;
  I: Integer;
  TreeView: TTreeView;
  Node: TTreeNode;
begin
  if csDestroying in Application.ComponentState then
    Exit;

  MessageViewForm := FindMessageViewForm;
  if Assigned(MessageViewForm) then
  begin
    if ScrollToLast then
    begin
      TreeView := nil;
      with MessageViewForm do
        for I := 0 to ControlCount - 1 do
          if Controls[I].ClassNameIs('TTreeMessageView') then
          begin
            TreeView := Controls[I] as TTreeView;
            Break;
          end;

      if Assigned(TreeView) then
      begin
        Node := TreeView.Items.GetFirstNode;
        while Assigned(Node) and (Node.GetNextSibling <> nil) do
          Node := Node.GetNextSibling;
        if Assigned(Node) then
          Node.MakeVisible;
      end;
    end;

    MessageViewForm.Show;
  end;
end;

function PathGetShortName(const APath: string): string;
var
  Buffer: array[0..MAX_PATH] of Char;
  Required: Integer;
begin
  Required := GetShortPathName(PChar(APath), Buffer, SizeOf(Buffer));
  if Required > MAX_PATH then // Buffer to small
  begin
    SetLength(Result, Required - 1);
    GetShortPathName(PChar(APath), Pointer(Result), Required);
  end
  else if Required = 0 then // Error
    Result := APath
  else
    SetString(Result, Buffer, Required);
end;

function TCrossKylixAddIn.GetIDString: string;
begin
  Result := 'scamp.CrossKylix';
end;

function TCrossKylixAddIn.GetName: string;
begin
  Result := 'CrossKylix';
end;

function TCrossKylixAddIn.GetState: TWizardState;
begin
  Result := [wsEnabled];
end;

procedure Register;
begin
  RegisterPackageWizard(TCrossKylixAddIn.Create as IOTAWizard)
end;

constructor TCrossKylixAddIn.Create;
var
  i: integer;
begin
  inherited;
  ProjectMenu := nil;
  ToolsMenu := nil;
  SeparatorMenuItem := nil;
  CompileMenuItem := nil;
  BuildMenuItem := nil;
  ToolsMenuItem := nil;
  with TMenuItem((BorlandIDEServices as INTAServices).MainMenu.items) do
  begin
    for i := 0 to count - 1 do
    begin
      if items[i].name = 'ProjectMenu' then
      begin
        ProjectMenu := items[i];
      end
      else if items[i].name = 'ToolsMenu' then
      begin
        ToolsMenu := items[i];
      end;
    end;
  end;
  if assigned(ProjectMenu) then
  begin
    SeparatorMenuItem := TMenuItem.Create(projectMenu);
    SeparatorMenuItem.Caption := '-';
    projectMenu.Add(SeparatorMenuItem);
    CompileMenuItem := TMenuItem.Create(projectMenu);
    CompileMenuItem.Caption := 'Compile with CrossKylix';
    CompileMenuItem.OnClick := compilekylix;
    projectMenu.Add(CompileMenuItem);
    BuildMenuItem := TMenuItem.Create(projectMenu);
    BuildMenuItem.Caption := 'Build with CrossKylix';
    BuildMenuItem.OnClick := buildkylix;
    projectMenu.Add(BuildMenuItem);
  end;
  if assigned(ToolsMenu) then
  begin
    ToolsMenuItem := TMenuItem.Create(ToolsMenu);
    ToolsMenuItem.Caption := 'CrossKylix Options...';
    ToolsMenuItem.OnClick := showoptions;
    ToolsMenu.Insert(4, ToolsMenuItem);
  end;

  cd := TCompileDialog.Create(application.mainform);

  od := TOptionsDialog.Create(application.mainform);
  ini := TMemInifile.create(extractfilepath(paramstr(0)) + 'crosskylix.ini');
  basedir := ini.ReadString('CrossKylix', 'BaseDir', '');
  if basedir <> '' then basedir := IncludeTrailingPathDelimiter(basedir);
  if pos(' ', basedir) > 0 then
  begin
    basedir := PathGetShortName(basedir);
  end;
  unitdirectories := ini.ReadString('CrossKylix', 'UnitDirectories', '');
  Encoding := ini.ReadString('CrossKylix', 'Encoding', 'ISO-8859-15');
  Includedirectories := ini.ReadString('CrossKylix', 'IncludeDirectories', '');
  Objectdirectories := ini.ReadString('CrossKylix', 'ObjectDirectories', '');
  debugoutput := ini.ReadBool('CrossKylix', 'DebugOutput', false);
  debugtrace := ini.ReadBool('CrossKylix', 'DebugTrace', false);
  UsePatchedLibc := ini.ReadBool('CrossKylix', 'UsePatchedLibc', true);

  if not fileexists(basedir + 'ck.dat') then
    showoptions(self);
end;

destructor TCrossKylixAddIn.Destroy;
begin
  //  ProjectMenu.Remove(CompileMenuItem);
  //  ProjectMenu.Remove(BuildMenuItem);
  try
    if assigned(CompileMenuItem) then freeandnil(CompileMenuItem);
    if assigned(BuildMenuItem) then freeandnil(BuildMenuItem);
    if assigned(ToolsMenuItem) then freeandnil(ToolsMenuItem);
    if assigned(SeparatorMenuItem) then freeandnil(SeparatorMenuItem);
    cd.Close;
    od.Close;
    freeandnil(cd);
    freeandnil(od);
    ini.free;
  finally
    inherited;
  end;
end;

procedure TCrossKylixAddIn.buildkylix(Sender: TObject);
begin
  docompile(true);
end;

procedure TCrossKylixAddIn.compilekylix(Sender: TObject);
begin
  docompile(false);
end;

procedure TCrossKylixAddIn.execute;
begin
  //
end;

function getlinuxpath(s: string): string;
var
  i, j: integer;
  sl: TStringList;
begin
  sl := TStringList.Create;

  while pos(';', s) > 0 do
  begin
    sl.add(copy(s, 1, pos(';', s) - 1));
    delete(s, 1, pos(';', s));
  end;
  if length(s) > 0 then sl.add(s);

  result := '';
  for i := 0 to sl.count - 1 do
  begin
    s := trim(sl[i]);
    if s = '' then continue;

    // if the path has got spaces in it, convert it to a short path.
    if pos(' ', s) > 0 then
    begin
      if directoryexists(s) then
        s := PathGetShortName(s);
    end;

    // drive letters in paths are pretty bad for linux, especially as the : is used as path delimiter...
    // so let's replace A:\-Z:\ and a:\-z:\ to /DRIVELETTER/
    for j := 65 to 90 do
    begin
      s := StringReplace(s, chr(j) + ':\', '/' + chr(j) + '/', [rfReplaceAll]);
    end;
    for j := 97 to 122 do
    begin
      s := StringReplace(s, chr(j) + ':\', '/' + chr(j) + '/', [rfReplaceAll]);
    end;

    // convert slashes
    s := StringReplace(s, '\', '/', [rfReplaceAll]);
    result := result + s;
    if i < sl.count - 1 then result := result + ':';
  end;
  sl.free;
end;

function getwindowspath(s: string): string;
begin
  if (length(s) > 3) and (s[1] = '/') and (s[3] = '/') then
  begin
    s[1] := s[2];
    s[2] := ':';
  end;
  s := StringReplace(s, '/', '\', [rfReplaceAll]);
  result := s;
end;

procedure TCrossKylixAddIn.docompile(build: boolean);
var
  i: integer;
  projectname, linuxprojectpath: string;

  flagfound: boolean;

  commandline: string;
  projectconfig: TStringlist;

  currentproject: IOTAProject;
  outputstream: TMemoryStream;
  outputfilename: string;
begin
  currentdir := getcurrentdir;

  currentproject := GetCurrentProject;
  if not assigned(currentproject) then exit;

  projectname := GetMainFilename(currentproject);

  projectpath := extractfilepath(projectname);
  if pos(' ', projectpath) > 0 then
  begin
    projectpath := PathGetShortName(projectpath);
  end;

  projectname := extractfilename(projectname);
  dolog('Project set to: ' + projectname);

  currentproject.save(false, false);

  linuxprojectpath := getlinuxpath(projectpath);
  linuxbasedir := getlinuxpath(basedir);

  (BorlandIDEServices as IOTAMessageServices).ClearCompilerMessages;

  ShowMessageView(false);
  cd.okButton.enabled := false;
  cd.label_project.caption := 'Project:' + #9#9 + projectpath + projectname;
  cd.label_status.caption := '';
  cd.label_status.font.Style := [];
  cd.label_errors.Caption := '0';
  cd.label_hints.caption := '0';
  cd.label_warnings.caption := '0';
  cd.Show;

  dolog('Writing configuration...');
  projectconfig := TStringList.Create;
  try
    with projectconfig do
    begin
      loadfromfile(projectpath + copy(projectname, 1, length(projectname) - length(ExtractFileExt(projectname))) + '.cfg');
      i := count - 1;
      repeat
        if (pos('-A', strings[i]) = 1) or (pos('-U', strings[i]) = 1) or (pos('-O', strings[i]) = 1) or (pos('-I', strings[i]) = 1) or (pos('-R', strings[i]) = 1) or
          (pos('-N', strings[i]) = 1) or (pos('-LE', strings[i]) = 1) or (pos('-LN', strings[i]) = 1) or (pos('-E', strings[i]) = 1) then
        begin
          strings[i] := getlinuxpath(strings[i]);
        end
        else if (pos('-w-UNSAFE_TYPE', strings[i]) = 1) or
          (pos('-w-UNSAFE_CODE', strings[i]) = 1) or (pos('-w-UNSAFE_CAST', strings[i]) = 1) then
          projectconfig.delete(i); // unsupported in kylix
        i := i - 1;
      until i < 0;

      // let's add some default paths.
      add('-U"' + getlinuxpath(UnitDirectories) + ':' + linuxprojectpath + '"');
      add('-I"' + getlinuxpath(IncludeDirectories) + ':' + linuxprojectpath + '"');
      add('-O"' + getlinuxpath(ObjectDirectories) + ':' + linuxprojectpath + '"');

      if UsePatchedLibc then
      begin
        // add replacement libc to config using a unit alias
        add('-U"' + linuxbasedir + 'libc' + '"');
        add('-ALibc=ckLibc');
      end;

      // If no exe output directory is defined, set it to project dir
      flagfound := false;
      for i := 0 to count - 1 do
      begin
        if pos('-E', strings[i]) = 1 then
        begin
          flagfound := true;
          break;
        end;
      end;
      if not flagfound then add('-E"' + linuxprojectpath + '"');

      // same for unit output dir
      flagfound := false;
      for i := 0 to count - 1 do
      begin
        if pos('-U', strings[i]) = 1 then
        begin
          flagfound := true;
          break;
        end;
      end;
      if not flagfound then add('-U"' + linuxprojectpath + '"');

      savetofile(projectpath + copy(projectname, 1, length(projectname) - length(ExtractFileExt(projectname))) + '.conf');
    end;
  except
    on e: exception do
    begin
      cd.hide;
      dolog('Error writing configuration: ' + e.message);
      MessageBox(0, 'CrossKylix was unable to create your projects Linux configuration file (.conf).' + #13 + #10 + 'To do this, it needs to access your projects Windows configuration file (.cfg).' + #13 + #10 + '' + #13 + #10 + 'The most common reason for this error is: You did not save your project' + #13 + #10 + 'in Delphi prior to compiling with CrossKylix.', 'CrossKylix', MB_ICONSTOP or MB_OK);
    end;
  end;

  w_count := 0;
  e_count := 0;
  h_count := 0;
  l_count := 0;
  cc := 0;
  if build then
  begin
    dolog('Building ' + projectname + '...');
    commandline := '-B ' + linuxprojectpath + projectname;
  end
  else
  begin
    dolog('Compiling ' + projectname + '...');
    commandline := linuxprojectpath + projectname;
  end;
  if debugoutput then dolog('Commandline: ' + commandline);

  outputstream:=TMemoryStream.Create;
  outputfilename:=compilewithdcc(basedir, commandline, encoding, msgoutput, outputstream, debugtrace);
  application.processmessages;

  if (outputfilename<>'') and (outputstream.size>0) then
  begin
    outputfilename:=getwindowspath(outputfilename);
    fixsegments(outputstream.memory, msgoutput, debugoutput);
    outputstream.SaveToFile(outputfilename);
    dolog('Created Linux binary in '+outputfilename);
  end
  else
  begin
    dolog('No Linux binary was created.');
  end;

  outputstream.free;

  l_count := l_count + cc;

  cd.label_total.caption := inttostr(l_count);
  cd.label_status.font.Style := [fsbold];

  if e_count > 0 then
    cd.label_status.caption := 'Done:' + #9 + 'There are errors.'
  else if w_count > 0 then
    cd.label_status.caption := 'Done:' + #9 + 'There are warnings.'
  else if h_count > 0 then
    cd.label_status.caption := 'Done:' + #9 + 'There are hints.'
  else
    cd.label_status.caption := 'Done.';
  ShowMessageView(true);
  cd.okButton.enabled := true;
end;

procedure dolog(const msg: string);
begin
  (BorlandIDEServices as IOTAMessageServices).AddToolMessage('',
    msg, 'CrossKylix', -1, -1);
  application.processmessages;
end;

procedure TCrossKylixAddIn.showoptions(Sender: TObject);
begin
  od.unit_directories.text := UnitDirectories;
  od.Include_directories.text := IncludeDirectories;
  od.Object_directories.text := ObjectDirectories;
  od.debugcheck.checked := DebugOutput;
  od.tracecheck.checked := DebugTrace;
  od.encodingcombo.ItemIndex := od.encodingcombo.Items.IndexOf(Encoding);

  if basedir <> '' then
    od.install_folder.text := basedir;
  od.install_folderChange(self);
  if od.showmodal = mrOk then
  begin
    basedir := IncludeTrailingPathDelimiter(od.install_folder.text);
    if pos(' ', basedir) > 0 then
    begin
      basedir := PathGetShortName(basedir);
    end;

    UnitDirectories := od.unit_directories.text;
    IncludeDirectories := od.Include_directories.text;
    ObjectDirectories := od.Object_directories.text;
    DebugOutput := od.debugcheck.checked;
    DebugTrace := od.tracecheck.checked;
    Encoding := od.encodingcombo.text;
    ini.WriteString('CrossKylix', 'BaseDir', basedir);
    ini.WriteString('CrossKylix', 'UnitDirectories', UnitDirectories);
    ini.WriteString('CrossKylix', 'ObjectDirectories', ObjectDirectories);
    ini.WriteString('CrossKylix', 'IncludeDirectories', IncludeDirectories);
    ini.WriteString('CrossKylix', 'Encoding', Encoding);
    ini.WriteBool('CrossKylix', 'DebugOutput', Debugoutput);
    ini.WriteBool('CrossKylix', 'DebugTrace', DebugTrace);
    ini.WriteBool('CrossKylix', 'UsePatchedLibc', UsePatchedLibc);
    ini.UpdateFile;
  end;
end;

procedure TCrossKylixAddIn.msgoutput(const s: string);
var
  i, j: integer;
  fname, efile, etmp: string;
  eline, ecol: integer;
  ekind: string;
  emsg: string;
begin
  Line := Line + s;
  while (pos(#13, line) > 0) or (pos(#10, line) > 0) do
  begin
    // check for CRLF, CR, LF
    i := pos(#13#10, line);
    j := pos(#13, line);
    if (i = 0) or ((j > 0) and (j < i)) then i := j;
    j := pos(#10, line);
    if (i = 0) or ((j > 0) and (j < i)) then i := j;

    if i > 0 then
    begin
      fname := trim(copy(line, 1, i));
      delete(line, 1, i);
      if debugoutput then dolog('DEBUG: ' + trim(fname));
      if fname <> '' then
      begin
        if (pos(' Fatal: ', fname) > 0) or (pos(' Error: ', fname) > 0) or (pos(' Warning: ', fname) > 0) or (pos(' Hint: ', fname) > 0) then
        begin
          efile := copy(fname, 1, pos('(', fname) - 1);
          efile := getwindowspath(efile);
          if (pos(':\', efile) < 2) then efile := projectpath + efile;

          delete(fname, 1, pos('(', fname));
          etmp := copy(fname, 1, pos(')', fname) - 1);
          delete(fname, 1, pos(')', fname) + 1);
          if pos(',', etmp) > 0 then
          begin
            eline := strtointdef(copy(etmp, 1, pos(',', etmp) - 1), -1);
            delete(etmp, 1, pos(',', etmp));
            ecol := strtointdef(etmp, 1);
          end
          else
          begin
            eline := strtointdef(etmp, -1);
            ecol := 1;
          end;
          ekind := copy(fname, 1, pos(':', fname) - 1);
          if ekind = 'Error' then
          begin
            e_count := e_count + 1;
            cd.label_errors.Caption := inttostr(e_count);
          end
          else if ekind = 'Fatal' then
          begin
            e_count := e_count + 1;
            cd.label_errors.Caption := inttostr(e_count);
          end
          else if ekind = 'Hint' then
          begin
            h_count := h_count + 1;
            cd.label_hints.Caption := inttostr(h_count);
          end
          else if ekind = 'Warning' then
          begin
            w_count := w_count + 1;
            cd.label_warnings.Caption := inttostr(w_count);
          end;

          delete(fname, 1, pos(':', fname) + 1);
          emsg := fname;
          (BorlandIDEServices as IOTAMessageServices).AddToolMessage(efile, emsg, ekind, eline, ecol);
          application.processmessages;
        end
        else
        begin
          if (pos('.pas', fname) > 0) or (pos('.dpr', fname) > 0) or (pos('.dpk', fname) > 0) then
          begin
            efile := copy(fname, 1, pos('(', fname) - 1);
            while pos('/', efile) > 0 do
              delete(efile, 1, 1);
            delete(fname, 1, pos('(', fname));
            etmp := copy(fname, 1, pos(')', fname) - 1);
            if currentfile <> efile then
            begin
              l_count := l_count + cc;
              cd.label_total.caption := inttostr(l_count);
            end;

            cc := strtointdef(etmp, 0);

            cd.label_line.caption := inttostr(cc);
            if currentfile <> efile then
            begin
              currentfile := efile;
              cd.label_status.caption := 'Compiling:' + #9 + efile;
            end;
          end
          else if (pos('Fatal', fname) > 0) or (pos('Error', fname) > 0) or (pos('Warning', fname) > 0) then
          begin
            if (pos('Fatal', fname) > 0) or (pos('Error', fname) > 0) then
            begin
              e_count := e_count + 1;
              cd.label_errors.Caption := inttostr(e_count);
              dolog(fname);
            end
            else if (pos('Warning', fname) > 0) and (pos('libX11.so.6', fname) = 0) then
            begin // ignore libX11.so.6 warning, it doesn't matter.
              w_count := w_count + 1;
              cd.label_warnings.Caption := inttostr(w_count);
              dolog(fname);
            end
          end;
        end;
      end;
    end;
  end;
end;

end.

